"""Abstract Widget Dependency."""
from typing import Any
from typing import Generator
from typing import Optional
from typing import Tuple

from engineai.sdk.dashboard.utils import generate_input

from .interface import DependencyInterface


class AbstractWidgetDependency(DependencyInterface):
    """Specs base for defining the Widget Dependency."""

    API_DEPENDENCY_INPUT: Optional[str] = None

    def __init__(self, *, dependency_id: str, widget_id: str, path: str):
        """Creates dependency with a widget.

        Args:
            dependency_id (str): id of dependency (to be used in other dependencies)
            widget_id (str): id of widget to associate dependency with
            path (str): path for state exposed by widget
        """
        super().__init__()
        self.__dependency_id = dependency_id
        self.__widget_id = widget_id
        self.__path = path

    def __iter__(self) -> Generator[Tuple[str, str], None, None]:
        yield "dependency_id", self.__dependency_id
        yield "widget_id", self.__widget_id
        yield "path", self.__path

    def __hash__(self) -> int:
        return hash(tuple(self))

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, type(self)) and tuple(self) == tuple(other)

    @property
    def api_dependency_input(self) -> str:
        """Return the API input to cast in api types during build."""
        if self.API_DEPENDENCY_INPUT is None:
            raise NotImplementedError(
                f"Class {self.__class__.__name__}.API_DEPENDENCY_INPUT not defined."
            )
        return self.API_DEPENDENCY_INPUT

    @property
    def dependency_id(self) -> str:
        """Return Dependency ID."""
        return self.__dependency_id

    @property
    def widget_id(self) -> str:
        """Returns id of widget associated with WidgetDependency.

        Returns:
            str: widget id
        """
        return self.__widget_id

    @widget_id.setter
    def widget_id(self, widget_id: str) -> None:
        """Change widget_id associated with WidgetDependency."""
        self.__widget_id = widget_id

    def build(self) -> Any:
        """Builds spec for dashboard API.

        Returns:
            Any: Input object for Dashboard API
        """
        return generate_input(
            self.api_dependency_input,
            widget=generate_input(
                "WidgetDependencyInput",
                widgetId=self.__widget_id,
                path=self.__path,
                name=self.__dependency_id,
            ),
        )
