"""Spec for a area series of a Timeseries widget."""
import re
import warnings
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union

import pandas as pd
from beartype import beartype

from engineai.sdk.dashboard.links.typing import GenericLink
from engineai.sdk.dashboard.styling.color.palette import qualitative_palette
from engineai.sdk.dashboard.styling.color.typing import ColorSpec
from engineai.sdk.dashboard.templated_string import TemplatedStringItem
from engineai.sdk.dashboard.templated_string import build_templated_strings
from engineai.sdk.dashboard.utils import generate_input
from engineai.sdk.dashboard.widgets.components.charts.series.base import ChartSeriesBase
from engineai.sdk.dashboard.widgets.components.charts.series.entities.typing import (
    Entities,
)
from engineai.sdk.dashboard.widgets.components.charts.styling.typing import (
    ColoredSeriesStyling,
)
from engineai.sdk.dashboard.widgets.components.charts.styling.typing import (
    SeriesStyling,
)
from engineai.sdk.dashboard.widgets.components.charts.tooltip.item import (
    build_tooltip_item,
)

from ...components.charts.typing import TooltipItem
from ...components.charts.typing import TooltipItems
from ...utils import get_tooltips
from ..exceptions import TimeseriesTooltipsEmptyDefinitionError
from ..exceptions import TimeseriesValidateSeriesDataColumnNotFound
from ..transform import Transform


class TimeseriesBaseSeries(ChartSeriesBase):
    """Spec for a Base Series of a Timeseries widget."""

    _styling_class: Optional[Type[ColoredSeriesStyling]] = None
    _INPUT_KEY: Optional[str] = None

    @beartype
    def __init__(
        self,
        *,
        name: Optional[Union[str, GenericLink]] = None,
        data_column: Optional[Union[str, GenericLink]],
        styling: Optional[SeriesStyling] = None,
        entity: Optional[Entities] = None,
        show_in_legend: bool = True,
        required: bool = True,
        visible: bool = True,
        right_axis: bool = False,
        tooltips: Optional[TooltipItems] = None,
    ):
        """Construct area series.

        Args:
            name: series name (shown in legend and tooltip).
            data_column: name of column in pandas dataframe(s) used for the values of
                this series.
            styling: styling spec.
                Defaults to None, i.e. picks color automatically from palette.
            entity: entity spec.
            show_in_legend: whether to show series in legend or not.
            required: Flag to make the Series mandatory. If required == True and no
                Data the widget will show an error. If required==False and no Data,
                the widget hides the Series.
            visible: Flag to make the Series visible when chart is loaded.
            right_axis: Flag to make the Series visible on the right axis. Only works
                when used in set_series, otherwise ignored.
            tooltips: tooltip items to be displayed at Series level.
        """
        super().__init__(name=name, data_column=data_column)
        self._data_column = data_column
        self._show_in_legend = show_in_legend
        self._tooltip_items = get_tooltips(tooltips)
        self._required = required
        self._visible = visible
        self._styling = styling
        self._entity = entity
        self._transforms: List[Transform] = []
        self._date_column: TemplatedStringItem = " "
        self.__is_right_axis = right_axis

    @property
    def is_right_axis(self) -> bool:
        """Flag to make the Series visible on the right axis."""
        return self.__is_right_axis

    @property
    def _input_key(self) -> str:
        """Returns styling Input Key argument value."""
        if self._INPUT_KEY is None:
            raise NotImplementedError(
                f"Class {self.__class__.__name__}._INPUT_KEY not defined."
            )
        return self._INPUT_KEY

    def add_tooltips(self, *tooltips: TooltipItem) -> "TimeseriesBaseSeries":
        """Add extra tooltip items (i.e. in addition to the value of this series)."""
        warnings.warn(
            "`add_tooltips` is deprecated. Use `tooltips` in constructor, instead.",
            DeprecationWarning,
        )
        if len(tooltips) == 0:
            raise TimeseriesTooltipsEmptyDefinitionError()

        self._tooltip_items.extend(tooltips)
        return self

    def add_transforms(self, *transform: Transform) -> "TimeseriesBaseSeries":
        """Add transform (i.e. apply transform to series data with low column data).

        Returns:
            AreaSeries: reference to this series to facilitate inline
                manipulation
        """
        self._transforms.extend(transform)
        return self

    def prepare(self, date_column: TemplatedStringItem, index: int) -> None:
        """Prepare series to be rendered."""
        self._date_column = date_column
        if self._styling is None:
            self.__set_default_styling(color=qualitative_palette(index=index))
        else:
            self._styling.prepare(self._data_column)

    def validate(self, *, data: pd.DataFrame) -> None:
        """Validate if dataframe that will be used for series contains required columns.

        Args:
            data: pandas dataframe which will be used for table
        """
        self._validate_data_column(
            data=data, data_column=self._data_column, data_column_name="data_column"
        )

        for item in self._tooltip_items:
            item.validate(data=data)

        if self._styling is not None:
            self._styling.validate(data=data)

    def build(self) -> Any:
        """Build series Input spec."""
        return generate_input(
            "TimeseriesWidgetSeriesInput", **{self._input_key: self._build_series()}
        )

    def _validate_data_column(
        self,
        *,
        data: pd.DataFrame,
        data_column: Optional[Union[str, GenericLink]],
        data_column_name: str,
    ) -> None:
        if data_column is not None and isinstance(data_column, str):
            # TODO find a way to validate templated strings
            if not re.search("{{(.*?)}}", data_column):
                if data_column not in data.columns:
                    raise TimeseriesValidateSeriesDataColumnNotFound(
                        series_class_name=self.__class__.__name__,
                        column_name=data_column_name,
                        column_value=data_column,
                    )

    def _build_series(self) -> Any:
        """Method implemented by all factories to generate Input spec.

        Returns:
            Input object for Dashboard API
        """
        return generate_input(
            self._api_type,
            name=build_templated_strings(items=self.name),
            xAxisKey=build_templated_strings(items=self._date_column),
            showInLegend=self._show_in_legend,
            tooltipItems=[
                build_tooltip_item(item=item) for item in self._tooltip_items
            ],
            required=self._required,
            isVisible=self._visible,
            **{"yAxisKey": build_templated_strings(items=self._data_column)}
            if self._data_column is not None
            else {},
            **self._build_extra_inputs(),
            styling=self._styling.build() if self._styling is not None else None,
            entity=self._entity.build() if self._entity is not None else None,
        )

    def _build_extra_inputs(self) -> Dict[str, Any]:
        return {}

    def __set_default_styling(self, color: ColorSpec) -> None:
        if self._styling_class is None:
            raise NotImplementedError(
                f"Class {self.__class__.__name__} does not have the variable "
                f"`_styling_class` initialized."
            )
        self._styling = self.__create_styling(self._styling_class, color=color)

    def __create_styling(
        self, styling_class: Type[ColoredSeriesStyling], color: ColorSpec
    ) -> ColoredSeriesStyling:
        return styling_class(color_spec=color)
